#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

# #  (Discrete) Truncated Gaussian distribution
#  Author: wenchieh
#
#  Project: eaglemine
#      truncated_gaussian.py
#      Version:  1.0
#      Date: December 17 2017
#      Main Contact: Wenchieh Feng (wenchiehfeng.us@gmail.com)
#
#      Copyright:
#        This software is free of charge under research purposes.
#        For commercial purposes, please contact the author.
#
#      Created by @wenchieh  on <12/17/2017>
#

__author__ = 'wenchieh'

# third-party lib
import numpy as np
from numpy.linalg import det, inv

# project
from .norm_extras import mvnormcdf

class TruncatedGaussian(object):
    def __init__(self, lower_bound, upper_bound):
        self._ndim = 0
        self._mu, self._cov = None, None
        self._lower, self._upper = np.array(lower_bound), np.array(upper_bound)

    def set_para(self, mu, cov):
        self._mu = np.array(mu)
        self._ndim = len(mu)
        self._cov = np.array(cov)

    def _phi_(self, xs, mu, cov):
        nc = 2.0 * np.pi * np.sqrt(det(cov))
        norm_vect = np.exp(-1 * (xs - mu).dot(inv(cov)).dot((xs - mu).transpose()) / 2.0) / nc
        return norm_vect

    def _log_phi_(self, xs, mu, cov):
        return np.log(self._phi_(xs, mu, cov))

    def _psi_(self, xs, mu, cov, lower, upper):
        for k in range(self._ndim):
            if xs[k] < lower[k] or xs[k] > upper[k]:
                return 0

        p_xy = self._phi_(xs, mu, cov)
        normalizer = mvnormcdf(upper, mu, cov, lower)
        return p_xy * 1.0 / normalizer

    def _log_psi_(self, xs, mu, cov, lower, upper):
        return np.log(self._psi_(xs, mu, cov, lower, upper))

    def pdf(self, xs):
        ''' the pdf of xs in truncate normal distribution'''
        return self._psi_(xs, self._mu, self._cov, self._lower, self._upper)

    def cdf(self, xs):
        ''' the probability of xs in truncate normal distribution. given mu, Sigma, and truncated lower bound.'''
        for k in range(self._ndim):
            if xs[k] < self._lower[k] or xs[k] > self._upper[k]:
                return 0

        xs_upper = np.array(xs)
        return mvnormcdf(xs_upper, self._mu, self._cov, self._lower)

    def range_cdf(self, left, right):
        for k in range(self._ndim):
            if left[k] < self._lower[k] or right[k] > self._upper[k]:
                return 0

        denorm = 1.0 * mvnormcdf(self._upper, self._mu, self._cov, self._lower)
        if denorm == 0:
            return 0
        p = 1.0 * mvnormcdf(right, self._mu, self._cov, left)
        return np.max([0,  p / denorm])


#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

# #  Some basic mdl (model description length) tools
#  Author: wenchieh
#
#  Project: eaglemine
#      mdlbase.py
#      Version:  1.0
#      Date: December 10 2017
#      Main Contact: Wenchieh Feng (wenchiehfeng.us@gmail.com)
#
#      Copyright:
#        This software is free of charge under research purposes.
#        For commercial purposes, please contact the author.
#
#      Created by @wenchieh  on <12/10/2017>
# 

__author__ = 'wenchieh'

# third-party lib
import numpy as np

class MDLBase(object):
    FLOATBIT = 4 * 8     # # reference to: KDD'14  AutoPlait

    @staticmethod
    def float_mdl(val, schema=1, precision=10):
        # 1. using the fixed code-length to encoding a float
        # 2. proportional to the magnitude of the value and the given precision \eta
        code_length = 0.0
        if schema == 1:
           code_length = MDLBase.FLOATBIT
        elif schema == 2:
            sign_code = 1
            val = abs(val)
            val_code = np.log2(val * 10**(precision)) if val > 0 else precision
            code_length = sign_code + val_code

        return code_length

    @staticmethod
    def integer_mdl(z):
        """
        encoded size of an integer >=1 as by Peter Elias's 1975 Elias code for integers
        Elias code variant for encoding zero.
        :param z: non-negative integer
        :return: encode length
        """
        if z < 0:
            ValueError("The Elias encoding can't cope with the negative integer. "
                       "Input non-negative integer value.")

        z += 1   #to encode 0, here we define the code-schema all value add 1
        c = 0
        while z > 1:
            t = np.floor(np.log2(z))
            c += 1 + int(t)
            z = t
        c += 1    # here 1 is the halt bit for encoding.
        return c

    @staticmethod
    def integer_elias_decode(code):
        if len(code) <= 0:
            ValueError("Code input error!")

        z = 0
        if len(code) == 1:
            z = 1
        else:
            code = code[:-1]
            sig_bits = 2
            while len(code) > 0:
                k = int(code[:sig_bits], 2)
                if sig_bits >= len(code):
                    z = k
                    break
                else:
                    code = code[sig_bits:]
                    sig_bits = k + 1
                    continue
        z -= 1   # encode all non-negative number.
        return z

    @staticmethod
    def integer_elias_encode(z):
        if z < 0:
            ValueError("The Elias encoding can't cope with the negative integer. "
                       "Input non-negative integer value.")
        z += 1
        code = ''
        val_code = bin(z)[2:]
        while len(val_code) > 1:
            code = val_code + code
            pre_len = len(val_code)
            val_code = bin(pre_len - 1)[2:]
        code += '0'    # the halt bit.
        return code

    @staticmethod
    def seq_diff_mdl(p, q):
        """
        encode the difference between two integer sequences.
        :param p: [int, array] the actual values
        :param q: [int, array] the expect values
        :return:
        """
        p, q = np.round(p), np.round(q)

        ## deprecated
        ## approximate Elias code (not the shortest code length)
        # code_len = np.sum([1 + 2 * np.ceil(np.log2(d)) if d > 0 else 1 for d in np.abs(p - q)])

        # absolute difference encode
        # code_len = np.sum(np.abs(p - q)) + len(p)

        ## here 1 is the sign bit, integer_mdl(d) is the code-length of absolute-difference.
        code_len = np.sum([1 + MDLBase.integer_mdl(d) for d in np.abs(p - q)])
        return code_len


import numpy as np
from .._model import DMmodel
import scipy.sparse.linalg as slin
import numpy as np
from scipy.sparse import csc_matrix, coo_matrix, csr_matrix, lil_matrix


class EigenSpokes(DMmodel):
    def __init__(self, data_mat):
        self.data = data_mat

    def run(self, k = 10):
        sparse_matrix = self.data.to_scipy()
        sparse_matrix = sparse_matrix.asfptype()
        RU, RS, RVt = slin.svds(sparse_matrix, k)
        RV = np.transpose(RVt)
        U, S, V = np.flip(RU, axis=1), np.flip(RS), np.flip(RV, axis=1)

        n_row = U.shape[0]
        n_col = V.shape[0]

        x_lower_bound = -1 / np.sqrt(n_col + 1)
        y_lower_bound = -1 / np.sqrt(n_col + 1)
        x_upper_bound = 1 / np.sqrt(n_col + 1)
        y_upper_bound = 1 / np.sqrt(n_col + 1)

        real_index1 = S.shape[0] - 1
        real_index2 = S.shape[0] - 2

        x = U[:, real_index1]
        y = U[:, real_index2]
    

        list_x_lower_outliers = [index for index in range(len(x)) if x[index] > x_lower_bound]
        list_y_lower_outliers = [index for index in range(len(y)) if y[index] > y_lower_bound]
        list_x_upper_outliers = [index for index in range(len(x)) if x[index] < x_upper_bound]
        list_y_upper_outliers = [index for index in range(len(y)) if y[index] < y_upper_bound]

        outliers_index = list(set(list_x_lower_outliers) & set(list_y_lower_outliers) &
                              set(list_x_upper_outliers) & set(list_y_upper_outliers))
        inliers_index = list(set(range(len(x))).difference(outliers_index))

        outliers_index, inliers_index = inliers_index, outliers_index
        print("Outliters:")
        print(outliers_index)
        return outliers_index
    
    
    def anomaly_detection(self):
        return self.run()

    def save(self, outpath):
        pass



from gensim import similarities
import os, logging
import numpy as np
import pandas as pd
import gc

class NeighborSampler():
    def __init__(self, chip_name, stage = 0, x_data=None):
        self.index = None
        self.chip_name = chip_name
        self.stage = stage
        self.build(x_data)

    def build(self, x_data=np.array([])):
        corpus = []
        for i in range(x_data.shape[0]):
            gensim_format_vec = []
            for j in range(x_data.shape[1]):
                gensim_format_vec.append((j, x_data[i][j]))
            corpus.append(gensim_format_vec)
        logging.info("#sample to build index: %s" % x_data.shape[0])
        self.get_index(corpus,x_data.shape[1])

    def get_index(self, corpus=None, n_feature=None):
        self.index = similarities.MatrixSimilarity(corpus, num_features=n_feature)
        return self.index

    def get_topk(self, vec, k):
        gensim_format_vec = []
        for i in range(len(vec)):
            gensim_format_vec.append((i, vec[i]))
        sims = self.index[gensim_format_vec]

        sim_sort = sorted(list(enumerate(sims)), key=lambda item: item[1])
        top_k = sim_sort[0:k]
        return top_k


def load_gene_file(filapath, stage_index, output_filepath_list):
    list_data = []
    with open(filapath) as fp:
        next(fp) # skip header
        for line in fp:
            data = line.split("\t")[1:]
            x_data = np.zeros(len(data))
            for i in range(len(data)):
                x_data[i] = float(data[i])
            list_data.append(x_data)
            del data
            gc.collect() # the line has more than 800K element
    x_data = np.array(list_data)

    if stage_index:
        print(stage_index)
        # stage_index = stage_index.split(",")
        stage_index = [int(x) for x in stage_index]
        stage_index = np.array(stage_index)
    else:
        stage_index = np.array([0] * x_data.shape[0])
    n_stage = np.max(stage_index) + 1
    # print("gene file has " + str(n_stage) + " stages")
    for j in range(n_stage):
        idx = stage_index == j
        x_output = x_data[idx]
        np.savetxt(output_filepath_list[j], x_output, delimiter="\t")

def build_sampled_coexpression_matrix(x_data, argsoutput1, k = 1000, use_peason = False):
    logging.info("Co-expression matrix building process starts.")
    logging.info("data shape: %s %s" % (x_data.shape[0], x_data.shape[1]))
    n_gene = x_data.shape[0]

    neighbor_sampler = NeighborSampler("out", x_data = x_data)
    with open(argsoutput1, "w") as f:
        for i in range(n_gene):
            list_neighbor = neighbor_sampler.get_topk(x_data[i], k)
            list_neighbor.extend(neighbor_sampler.get_topk(x_data[i], k))
            list_neighbor = list(set(list_neighbor))
            logging.info("sample %s's topk neighbor: %s" % (i, list_neighbor))
            for j, value in list_neighbor:
                if use_peason:
                    value = pearson(x_data[i], x_data[j])
                f.write("%s\t%s\t%s\n" % (i, j, value))
            f.flush()
            os.fsync(f.fileno())

def get_top_by_threshold(input_path, output_path, threshold = 0.99):
    logging.info("start getting top by threshold")

    with open(output_path, "w") as wfp:
        with open(input_path) as fp:
            for line in fp:
                tokens = line.split()
                if len(tokens) != 3:
                    continue
                if threshold <= float(tokens[2]):
                    wfp.write("%s\t%s\n" % (tokens[0], tokens[1]))

def pearson(X, Y):
    return np.corrcoef(X, Y)[0][1]

def graph1_minus_graph2(g1, g2, only_pos = 0, output_path = "res_net.edgelist"):
    """
    If want to get edges in g1 and not in g2, only_pos = -1
    If want to get edges in g2 and not in g1, only_pos = 1
    """
    edge_list = []
    nodes1 = g1.nodes()
    nodes2 = set(g2.nodes())
    for node in nodes1:
        neighbors2 = set()
        neighbors1 = set(g1.neighbors(node))
        if node in nodes2:
            neighbors2 = set(g2.neighbors(node))

    if only_pos != 1:
        for neighbor in neighbors1:
            if neighbor not in neighbors2:
                edge_list.append((node, neighbor))
    if only_pos != -1:
        for neighbor in neighbors2:
            if neighbor not in neighbors1:
                edge_list.append((node, neighbor))

    with open(output_path, "w") as wfp:
        for edge in edge_list:
            wfp.write(str(edge[0])+"\t"+str(edge[1])+"\n")
    return edge_list

if __name__ == "__main__":
    x_data = pd.read_csv("F:\gene2\spartan2\live-tutorials\inputData\example.csv", sep=",", header=None)
    build_sampled_coexpression_matrix(x_data, "out")


import numpy as np
from .._model import DMmodel
from . import param_default

param_default_dict = {
    'signal_freq': 360,
    'Smin': 180,
    'Smax': 300,
    'max_dist': 250,
    'prediction_length': 0,
    'model_momentum': 0,
    'max_vocab': 100,
    'termination_threshold': 0,
    'new_cluster_threshold': 0.3
}


class BeatLex(DMmodel):
    def __init__(self, data_mat, model_name='my_beatlex_model', *args, **para_dict):
        super(BeatLex, self).__init__(data_mat, model_name, *args, **para_dict)
        self.data_mat = data_mat.val_tensor._data

        self.signal_freq = param_default(para_dict, 'signal_freq', param_default_dict)
        self.Smin = param_default(para_dict, 'Smin', param_default_dict)
        self.Smax = param_default(para_dict, 'Smax', param_default_dict)
        self.max_dist = param_default(para_dict, 'max_dist', param_default_dict)
        self.prediction_length = param_default(para_dict, 'prediction_length', param_default_dict)

        self.model_momentum = param_default(para_dict, 'model_momentum', param_default_dict)
        self.max_vocab = param_default(para_dict, 'max_vocab', param_default_dict)
        self.termination_threshold = param_default(para_dict, 'termination_threshold', param_default_dict)
        self.new_cluster_threshold = param_default(para_dict, 'new_cluster_threshold', param_default_dict)

        if 'models' in para_dict.keys():
            self.models = para_dict['models']
        else:
            self.models = []

    def _summarize_sequence(self):
        if len(self.models) == 0:
            start_init_pos = 0
            best_place = self.get_new_segement_size(0)
            end_init_pos = best_place[0] - 1
            print(f'init at {end_init_pos}')
            start_pos_list = [start_init_pos]
            end_pos_list = [end_init_pos]
            self.models.append(self.data_mat[:, start_pos_list[0]:end_pos_list[0]+1])
            idx = [0]
        else:
            start_pos_list = [-1]
            end_pos_list = [-1]
            idx = [-1]
        mean_dev = np.mean((self.data_mat[:] - np.mean(self.data_mat[:])) ** 2)
        best_prefix_length = np.nan
        tot_err = 0
        while end_pos_list[-1] + self.termination_threshold < self.data_mat.shape[1]:
            cur_idx = len(start_pos_list) + 1
            cur_pos = end_pos_list[-1] + 1
            start_pos_list.append(cur_pos)

            num_models = len(self.models)
            ave_costs = np.full((num_models, self.Smax), np.nan)
            cur_end = min(cur_pos + self.Smax - 1, self.data_mat.shape[1])
            if cur_pos == cur_end:
                break
            print('\n========segment {} at {} {}'.format(cur_idx, cur_pos, cur_end))
            data_cur = self.data_mat[:, cur_pos:cur_end]
            for k in range(num_models):
                dtw_dist, dtw_mat, dtw_k, dtw_ways = self.dynamic_time_warping(self.models[k], data_cur)
                dtw_costs = dtw_mat[-1, :]
                ave_costs[k, 0:data_cur.shape[1]] = dtw_costs / np.arange(1, data_cur.shape[1]+1, 1)
                ave_costs[k, 0:self.Smin-1] = np.nan
            best_cost = np.nanmin(ave_costs)
            min_place = np.where(ave_costs == best_cost)
            try:
                min_k = min_place[0][0]
                best_size = min_place[1][0]
            except:
                print('Sequence End Reached')
            if cur_pos + self.Smax >= self.data_mat.shape[1]:
                good_prefix_costs = np.full(num_models, np.nan)
                good_prefix_lengths = np.full(num_models, np.nan)
                for k in range(num_models):
                    _dtw_dist, _dtw_mat, _dtw_k, _dtw_ways = self.dynamic_time_warping(self.models[k], data_cur)
                    prefix_costs = _dtw_mat[:, -1].transpose()
                    ave_prefix_costs = prefix_costs / np.arange(1, len(self.models[k][0])+1, 1)
                    good_prefix_costs[k] = min(ave_prefix_costs)
                    good_prefix_lengths[k] = np.where(ave_prefix_costs == min(ave_prefix_costs))[0]
                best_prefix_cost = min(good_prefix_costs)
                best_prefix_k = np.where(good_prefix_costs == min(good_prefix_costs))[0]
                best_prefix_length = good_prefix_lengths[best_prefix_k]
                print('best prefix found {} {} {}'.format(min_k, best_cost, best_prefix_cost))
                end_pos_list.append(self.data_mat.shape[1])
                if best_prefix_cost > best_cost and len(self.models) < self.max_vocab:
                    print('ending with new cluster')
                    self.models.append(np.array(self.data_mat[:, start_pos_list[-1]:end_pos_list[-1]]))
                    idx.append(num_models)
                    break
                else:
                    print('ending with prefix {}'.format(best_prefix_k))
                    idx.append(best_prefix_k[0])
                    break
            print('cluster cost {}'.format(ave_costs[:, best_size]))
            print('new cluster cost for {}: {}'.format(self.data_mat.shape[0], self.new_cluster_threshold * mean_dev * self.data_mat.shape[0]))
            print('size chosen: {}'.format(best_size+1))
            data_best = self.data_mat[:, cur_pos:cur_pos+best_size+1]
            if best_cost > self.new_cluster_threshold * mean_dev and len(self.models) < self.max_vocab:
                print('new_cluster')
                best_place = self.get_new_segement_size(cur_pos)
                print('best_place: {}'.format(best_place))
                best_S1 = best_place[0]
                print('best_S1: {}'.format(best_S1))
                end_pos_list.append(cur_pos+best_S1)
                idx.append(num_models)
                self.models.append(np.array(self.data_mat[:, start_pos_list[-1]:end_pos_list[-1]+1]))
                print('new cluster starts {} ends {}'.format(start_pos_list[-1], end_pos_list[-1]))
                tot_err += self.new_cluster_threshold * mean_dev * (best_S1 + 1)
            else:
                print('cluster {}'.format(min_k))
                end_pos_list.append(cur_pos+best_size)
                idx.append(min_k)
                tot_err += best_cost*best_size
                _dtw_dist, _dtw_mat, _dtw_k, _dtw_ways = self.dynamic_time_warping(self.models[min_k], data_best)
                trace_summed = np.zeros(self.models[min_k].shape)
                for t in range(0, len(_dtw_ways)):
                    trace_summed[:, _dtw_ways[t][0]] = trace_summed[:, _dtw_ways[t][0]] + data_best[:, _dtw_ways[t][1]]
                trace_keys = list(set([i[0] for i in _dtw_ways]))
                trace_dict = dict.fromkeys(trace_keys, 0)
                for i in _dtw_ways:
                    trace_dict[i[0]] += 1
                trace_counts = list(trace_dict.values())
                trace_aves = trace_summed / trace_counts
                origin_num = idx.count(min_k)
                if self.model_momentum == 0:
                    self.models[min_k] = (origin_num-1) / origin_num * self.models[min_k] + 1/origin_num * trace_aves
                else:
                    self.models[min_k] = self.model_momentum * self.models[min_k] + (1 - self.model_momentum) * trace_aves
        tot_err = tot_err / np.std(self.data_mat, ddof=1) ** 2 + (len(idx) - 1) * np.log2(len(self.data_mat)) + len(idx) * np.log2(len(self.models))
        result = {
            'signal_freq': self.signal_freq,
            'tot_err': tot_err,
            'starts': start_pos_list,
            'ends': end_pos_list,
            'idx': idx,
            'best_prefix_length': best_prefix_length,
            'models': self.models,
        }
        return result

    def get_new_segement_size(self, cur_pos):
        num_models = len(self.models)
        ave_costs = np.full((self.Smax, num_models + 1, self.Smax), np.inf)
        for S in range(self.Smin, self.Smax+1):
            if cur_pos + S >= self.data_mat.shape[1]:
                continue
            for k in range(num_models + 1):
                if k < num_models:
                    cur_model = self.models[k]
                elif k == num_models:
                    cur_model = self.data_mat[:, cur_pos:cur_pos+S]
                data_cur = self.data_mat[:, cur_pos+S:min(self.data_mat.shape[1], cur_pos+S+self.Smax)]
                dist, dtw_mat, ans_k, ans_w = self.dynamic_time_warping(cur_model, data_cur)
                dtw_costs = dtw_mat[-1, :]
                ave_costs[S-1, k, 0:data_cur.shape[1]] = [dtw_costs[index] / (index+1) for index in range(len(dtw_costs))]
                ave_costs[S-1, k, 0:self.Smin] = np.inf
        best_place = np.where(ave_costs == np.nanmin(ave_costs))
        best_place = [x[0] for x in best_place]
        return best_place

    def dynamic_time_warping(self, cur_model, data_seg):
        '''
        cur_model, model chosen from model_list
        data_seg, current handled data
        '''
        rows, N = cur_model.shape
        M = data_seg.shape[1]
        d = self.cal_distance(cur_model, data_seg) / rows
        D = np.full(d.shape, np.inf)
        D[0, 0] = d[0, 0]
        for n in range(1, N):
            D[n, 0] = d[n, 0] + D[n-1, 0]
        for m in range(1, M):
            D[0, m] = d[0, m] + D[0, m-1]
        encode_cost = 1
        mcost = encode_cost * np.std(cur_model, ddof=1) * np.log2(M)
        ncost = encode_cost * np.std(data_seg, ddof=1) * np.log2(N)
        for n in range(1, N):
            m_min = max(1, n-self.max_dist)
            m_max = min(M, n+self.max_dist+1)
            for m in range(m_min, m_max):
                D[n, m] = d[n, m] + min(min(D[n-1, m] + mcost, D[n-1, m-1]), D[n, m-1]+ncost)
        Dist = D[-1, -1]
        n = N-1
        m = M-1
        k = 1
        w = [[n, m]]
        while n + m != 0:
            if n == 0:
                m = m - 1
            elif m == 0:
                n = n - 1
            else:
                num_list = [D[n-1, m], D[n, m-1], D[n-1, m-1]]
                min_val = min(num_list)
                min_pos = num_list.index(min_val)
                if min_pos == 0:
                    n = n - 1
                elif min_pos == 1:
                    m = m - 1
                elif min_pos == 2:
                    n = n - 1
                    m = m - 1
            k = k + 1
            w.append([n, m])
        return Dist, D, k, w

    def cal_distance(self, a, b):
        if a.shape[0] != b.shape[0]:
            raise Exception('A and B should be of same dimensionality')

        ans = np.zeros(shape=(a.shape[1], b.shape[1]))

        for i in range(a.shape[0]):
            _a = np.matrix(a[i])
            _b = np.matrix(b[i])
            aa = np.power(_a, 2)
            bb = np.power(_b, 2)
            ab = np.dot(_a.transpose(), _b)

            plus1 = np.tile(aa.transpose(), (1, bb.shape[1]))
            plus2 = np.tile(bb, (aa.shape[1], 1))
            ans += np.abs(plus1 + plus2 - 2*ab)
        return ans

    def summarization(self, params=None):
        return self._summarize_sequence()

    def run(self):
        return self.summarization()



#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

#  Project: eaglemine
#  File: eaglemine_main.py
#  Goal: The main routine of eaglemine algorithm
#      Version:  1.0
#      Goal: Test script
#      Created by @wenchieh  on <12/17/2017>
#

__author__ = 'wenchieh'

# sys
import os
import time
import argparse

# third-party lib
import numpy as np

# project
from .utils.loader import Loader
from .eaglemine_model import EagleMineModel


VERBOSE=True
node2lab = 'nodelabel.out'
hcel2lab = 'hcel2label.out'

def load_hcel_weights(in_hist, in_hcel2avgfeat, mode=2, wtcol_index=1, sep=','):
    loader = Loader()
    _, _, hist_arr = loader.load_histogram(in_hist)

    nhcubes = len(hist_arr)
    hcube2index = dict(zip(map(tuple, hist_arr[:, :2]), range(nhcubes)))
    hcube_weight = np.zeros(nhcubes)  #np.empty((nhcubes, nfeat))
    with open(in_hcel2avgfeat, 'r') as fp:
        for line in fp.readlines():
            if line.startswith('#'):  continue
            tok = line.strip().split(sep)
            pos = tuple(map(int, tok[:mode]))
            hcube_weight[hcube2index[pos]] = float(tok[wtcol_index + mode - 1])
        fp.close()

    return hcube_weight


def eaglemine(in_hist, in_node2hcel, in_hcel2avgfeat, outs,
              strictness=3, desc_voc="dtmnorm", wt_featidx=1, mode=2, mix_comps=2):
    if VERBOSE:
        print("histogram: %s;  hode2pos: %s;  hpos_avgfeat: %s" % (in_hist, in_node2hcel, in_hcel2avgfeat))

    print("EagleMine algorithm")
    print("*****************")
    print("[0]. initialization and loading")
    eaglemodel = EagleMineModel(mode, mix_comps)
    eaglemodel.set_vocabulary(desc_voc)
    eaglemodel.load_histogram(in_hist)

    print("*****************")
    print("[1]. WaterLevelTree")
    start_tm = time.time()
    eaglemodel.leveltree_build(outs, prune_alpha=0.80, verbose=VERBOSE)
    end_tm1 = time.time()
    print("done @ {}".format(end_tm1 - start_tm))

    print("*****************")
    print("[2]. TreeExplore")
    eaglemodel.search(strictness=strictness, verbose=VERBOSE)
    # eaglemodel.save(outfd, '_search_res')
    eaglemodel.post_stitch(verbose=VERBOSE)
    end_tm2 = time.time()
    print("done @ {}".format(end_tm2 - end_tm1))

    if wt_featidx >= 0:
        print("*****************")
        print("[3]. node groups suspicious measure")
        histpos_avgdeg = load_hcel_weights(in_hist, in_hcel2avgfeat, mode, wt_featidx)
        eaglemodel.cluster_weighted_suspicious(histpos_avgdeg, verbose=VERBOSE)

    print("*****************")
    print("saving result")
    eaglemodel.graph_node_cluster(in_node2hcel, os.path.join(outs, node2lab), os.path.join(outs, hcel2lab))
    eaglemodel.cluster_histogram(verbose=VERBOSE)
    eaglemodel.save(outs)

    print('done @ {}'.format(end_tm2 - start_tm))


if __name__ == '__main__':
    path = '../output/'
    in_hist = 'histogram.out'
    in_node2hcel = 'node2hcel.out'
    in_hcel2avgfeat = 'hcel2avgfeat.out'
    desc_voc="dtmnorm"

    wtft = 1   # degree index
    mode, mix_comps = 2, 2
    strictness = 4

    eaglemine(path + in_hist, path + in_node2hcel, path + in_hcel2avgfeat, path,
              strictness, desc_voc, wt_featidx=wtft, mode=mode, mix_comps=mix_comps)




class UnionFind:
    """Union-Find data structure, provides union and find operations. Implemented with path compression optimization.
    """
    def __init__(self, N):
        """Init a N-element Union-Find with ids ranging from [0, N-1].

        Args:
            N (int): number of elements
        """
        self.N = N
        self.id = list(range(N))
        self.sizes = [1] * N

    def union(self, x, y):
        """Union two disjoint sets with id x and y respectively.

        Args:
            x (int): id of the first element
            y (int): id of the second element
        """
        rootx = self._root(x)
        rooty = self._root(y)
        if rootx == rooty:
            return

        if self.sizes[rootx] >= self.sizes[y]:
            self.id[rooty] = rootx
            self.sizes[rootx] += self.sizes[rooty]
        else:
            self.id[rootx] = rooty
            self.sizes[rooty] += self.sizes[rootx]

    def find(self, x, y):
        """Return if two elements belong to the same set.

        Args:
            x (int): id of the first element
            y (int): id of the second element

        Returns:
            [bool]: True if x and y belongs to the same set.
        """
        return self._root(x) == self._root(y)

    def _root(self, x):
        if self.id[x] == x:
            return x
        else:
            self.id[x] = self._root(self.id[x])
            return self.id[x]